#include <string.h>
#include "_afintrn.h"
#include <allegro/aintern.h>


static double _scale_x = 1;
static double _scale_y = 1;
static int _fill = TRUE;
static int _antialias = FALSE;

static SPLINE_FONT_DATA *_find_range(SPLINE_FONT_DATA *start, int c)
{
   SPLINE_FONT_DATA *s = start;

   while(s)
   {
      if (c >= s->start && c <= s->end)
         return s;

      s = s->next_range;
   }

   return NULL;
}



static int _render_char(BITMAP *bmp, SPLINE_FONT_DATA *dat, int ch, int x, int y, int color)
{
   int cache_valid = FALSE;
   int render_depth = bitmap_color_depth(bmp);
   int aa = (bitmap_color_depth(bmp) == 8) ? FALSE : _antialias;
   int cache_depth = aa ? 32  : render_depth;
   SPLINE_SHAPE **s;

   dat = _find_range(dat, ch);

   if (!dat)
      return 0;

   ch -= dat->start;

   s = dat->glyphs;



   if (s[ch])
   {
      if (_fill)
      {
         if (s[ch]->cached)
         {
            if ((s[ch]->cached->w != s[ch]->w * _scale_x)
                || (s[ch]->cached->h != s[ch]->h * _scale_y)
                || (bitmap_color_depth(s[ch]->cached) != cache_depth)
                )
            {
               destroy_bitmap(s[ch]->cached);
               s[ch]->cached = NULL;
            }
            else
            {
               cache_valid = TRUE;
            }

            if (s[ch]->cached_color != color)
               cache_valid = FALSE;
         }

         /* *no* else, since the above if may have deleted it */
         if (!s[ch]->cached)
         {
            if (aa)
               s[ch]->cached = create_bitmap_ex(32,s[ch]->w * _scale_x, s[ch]->h * _scale_y);
            else
               s[ch]->cached = create_bitmap_ex(render_depth,s[ch]->w * _scale_x, s[ch]->h * _scale_y);
         }

         s[ch]->cached_color = color;

         
         if (aa)
         {
            drawing_mode(DRAW_MODE_TRANS, NULL, 0,0);
            if (!cache_valid)
            {
               clear_to_color(s[ch]->cached, makecol32(getr_depth(render_depth,color), getg_depth(render_depth,color), getb_depth(render_depth,color)));
               set_write_alpha_blender();
               draw_shape(s[ch]->cached, 0, 0, -1, _scale_x, _scale_y, s[ch]);
            }

            set_alpha_blender();
            draw_trans_sprite(bmp, s[ch]->cached, x, y);
            solid_mode();
         }
         else
         {
            if (!cache_valid)
            {
               clear_to_color(s[ch]->cached,bitmap_mask_color(s[ch]->cached));
               draw_shape(s[ch]->cached, 0, 0, color, _scale_x, _scale_y, s[ch]);
            }
            draw_sprite(bmp, s[ch]->cached, x, y);
         }
      }
      else
         draw_shape_outline(bmp, x, y, color, _scale_x, _scale_y, s[ch]);
   }

   return s[ch] ? s[ch]->w * _scale_x : 0;
}

/* will crash when there is no glyph at all in the font */
static int _height_hook(SPLINE_FONT_DATA *data)
{
   SPLINE_SHAPE **s = data->glyphs;
   int i = 0;

   while(1)
   {
    if (s[i])
    {
      return (int)(s[i]->h * _scale_y);
    }
    i++;
   }

   return 0;
}



static int _char_width(SPLINE_FONT_DATA *data, int ch)
{



   SPLINE_FONT_DATA *s = _find_range(data, ch);

   if (!s && ch == 32)
      return (_height_hook(data) *_scale_x/ 2);


   if (s)
   {
    ch = ch - s->start;

    if (s->glyphs[ch])
    {
         return (int)(_scale_x * s->glyphs[ch]->w);
    }
   }
   return 0;
}


static SPLINE_FONT_DATA *_create_range(int base, int numchar)
{
   SPLINE_FONT_DATA *ret = malloc(sizeof(SPLINE_FONT_DATA));

   if (!ret)
      return NULL;

   ret->glyphs = malloc(sizeof(SPLINE_SHAPE *) * numchar);

   if (!ret->glyphs)
   {
      free(ret);
      return NULL;
   }
   
   ret->start = base;
   ret->end = base + numchar - 1;

   ret->next_range = NULL;


   return ret;
}



static int _dynamic_font_text_length(AL_CONST FONT *f, AL_CONST char *text)
{
   SPLINE_FONT_DATA *s = (SPLINE_FONT_DATA *)(f->data);
   AL_CONST char* p = text;
   int ch, w = 0;


    while( (ch = ugetxc(&p)) ) {
        w += _char_width(s, ch);
    }
    return w;
}

void _dynamic_font_render(AL_CONST FONT* f, AL_CONST char* text, int fg, int bg, BITMAP* bmp, int x, int y)
{
   SPLINE_FONT_DATA *s = (SPLINE_FONT_DATA *)(f->data);
   AL_CONST char* p = text;
   int ch;

   if (bg)
   {
      solid_mode();
      rectfill(bmp, x, y, x + text_length(f, text), y + text_height(f), bg);
   }

    while( (ch = ugetxc(&p)) ) {
        x += _render_char(bmp, s, ch, x, y, fg);
    }
}


static void _dynamic_font_destroy(FONT *f)
{
   SPLINE_FONT_DATA *s = (SPLINE_FONT_DATA *)f->data;
   SPLINE_FONT_DATA *next;
   int i;

   while(s)
   {
      next = s->next_range;
      for (i = s->start; i<=s->end;i++)
      {
         destroy_shape(s->glyphs[i - s->start]);
      }

      free(s);

      s = next;
   }


   free(f);
   
}

static int _dynamic_font_text_height(FONT const *f)
{
   SPLINE_FONT_DATA *s = (SPLINE_FONT_DATA *)f->data;
   
   return _height_hook(s);
}

static FONT_VTABLE _dynamic_font_vtable =
{
   _dynamic_font_text_height,
   NULL, // char_length
   _dynamic_font_text_length,
   NULL, //render_char
   _dynamic_font_render,
   _dynamic_font_destroy,
};

/* creates an empty dynamic font, the shapes still need to be added
 * the glyphs pointer is just an arry of NULL void pointers
 * it does set the vtable, so it won't crash when you use it with
 * textout, though you won't see anything
 */
FONT *create_dynamic_font(int base, int numchar)
{
   FONT *f = malloc(sizeof(FONT));
   
   if (!f)
       return NULL;


   f->data = _create_range(base, numchar);

   if (!f->data)
   {
      free(f);
      return NULL;
   }

   f->vtable = &_dynamic_font_vtable;

   f->height = 0;

   return f;
}


/*
text font format:
The font file is just a tetx file saved as an allegro compressed packfile
if you unpack it with allegro's pack command you can look into it


----start---------
AlDynFont  must start line
base       -font base
numchar    - number of chars
-------for each char
numspline
w
h
-------for each spline
x1 y1 x2 y2 x3 y3 x4 y4   control pts

 */

FONT *splif_load_font(char *filename)
{
   char buffer[1024];
   char *nextpart;
   FONT *f = NULL;
   SPLINE_FONT_DATA *sf = NULL;
   PACKFILE *in = pack_fopen(filename, "rp");
   int numchar, numspline,w,h;
   int base;
   int i,j,k;
   int pts[8];

   if (!in)
      return NULL;


   if (!pack_fgets(buffer, 1024, in))
   {
      pack_fclose(in);
      return NULL;
   }

   if (strncmp(buffer, "AlDynFont", 9))
   {
      pack_fclose(in);
      return NULL;
   }

   if (!pack_fgets(buffer, 1024, in))
   {
      pack_fclose(in);
      return NULL;
   }

   base = strtol(buffer, NULL, 0);

   if (!pack_fgets(buffer, 1024, in))
   {
      pack_fclose(in);
      return NULL;
   }

   numchar = strtol(buffer, NULL, 0);

   f = create_dynamic_font(base, numchar);


   if (!f)
   {
      pack_fclose(in);
      return NULL;
   }


   sf = (SPLINE_FONT_DATA *)(f->data);
   
   for (i=0;i<numchar;i++)
   {
     if (!pack_fgets(buffer, 1024, in))
     {
        pack_fclose(in);
        destroy_font(f);
        return NULL;
     }
     numspline = strtol(buffer, NULL, 0);

     if (!pack_fgets(buffer, 1024, in))
     {
        pack_fclose(in);
        destroy_font(f);
        return NULL;
     }
     w = strtol(buffer, NULL, 0);
     if (!pack_fgets(buffer, 1024, in))
     {
        pack_fclose(in);
        destroy_font(f);
        return NULL;
     }
     h = strtol(buffer, NULL, 0);

     if (numspline || w || h)
     {
         sf->glyphs[i] = create_shape(w,h,numspline);

         if (!sf->glyphs[i])
         {
            pack_fclose(in);
            destroy_font(f);
            return NULL;
         }
     }
     else
     {
         sf->glyphs[i] = NULL;
     }
     
     for (j=0;j<numspline;j++)
     {
         if (!pack_fgets(buffer, 1024, in))
         {
          pack_fclose(in);
          destroy_font(f);
          return NULL;
         }
         nextpart = buffer;
         for (k=0;k<8;k++)
         {
           pts[k] = strtol(nextpart, &nextpart, 0);
         }
         shape_add_spline(sf->glyphs[i], j, pts);
     }

   }


   pack_fclose(in);

   return f;
}

void splif_setscale(double scale_x, double scale_y)
{
   _scale_x = scale_x;
   _scale_y = scale_y;
}

void splif_setfill(int fill)
{
 _fill = fill;
}


void splif_setantialias(int antialias)
{
   _antialias = antialias;
   
}


