#include "_afintrn.h"



/* creates a shape, the splines still need to be filled in */
SPLINE_SHAPE *create_shape(int w, int h, int nr_of_splines)
{
   SPLINE_SHAPE *s = malloc(sizeof(SPLINE_SHAPE) + nr_of_splines * 8 * sizeof(int));


   if (!s)
       return NULL;

   s->w = w;
   s->h = h;
   s->nr_splines = nr_of_splines;
   s->cached = NULL;
   s->cached_color = -1;

   return s;
}

void shape_add_spline(SPLINE_SHAPE *s, int index, int *controlpoints)
{
   int i;
   ASSERT(s);
   ASSERT(index < s->nr_splines);
   ASSERT(controlpoints);


   for (i=0; i<8;i++)
       s->data[index][i] = controlpoints[i];
}


void shape_translate(SPLINE_SHAPE *s, int x, int y)
{

   int i,j;

   for (i=0;i<s->nr_splines;i++)
   {
      for (j=0;j<4;j++)
      {
         s->data[i][2*j] += x;
         s->data[i][2*j+1] += y;
      }
   }
}

void shape_scale(SPLINE_SHAPE *s, double sx, double sy)
{
   int i,j;

   for (i=0;i<s->nr_splines;i++)
   {
      for (j=0;j<4;j++)
      {
         s->data[i][2*j] *= sx;
         s->data[i][2*j+1] *= sy;
      }
   }

}

void destroy_shape(SPLINE_SHAPE *s)
{
   if (s->cached)
      destroy_bitmap(s->cached);

   free(s);
}


/* deletes splines min to amx (inclusive) from the
 * object. The actual memory is not freed,
 * just the nr_splines variables and the changed and the splines are shifted
 */
void shape_delete_splines(SPLINE_SHAPE *s, int min, int max)
{
    int i,j;

    ASSERT(max < s->nr_splines);
    ASSERT(min >=0);

    for (i=min;i+max - min + 1< s->nr_splines;i++)
    {
       for (j=0;j<8;j++)
           s->data[i][j] = s->data[i+max - min + 1][j];
    }


    s->nr_splines -= (max - min +1);
}
